/*
  (C)2019 izumi@ieee.org, all rights reserved
*/

#define GEISTER_C

#define DEBUG 0

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include "geister.h"
#include "geisterlib.h"
#include "geisterclient.h"

/*----------------------------------------------------------------*/
int GstSetGhost(struct GstGameStr *s){
  const char color[2]={GstRed, GstBlue};
  float w[2]={4.0, 4.0};
  int i,c;

  for (i=0;i<8;i++) {
    c=WeightedRandom(w,2);
    s->ghost[GstPly0][i].color=color[c];
    w[c]-=1.0;
  }

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstPlay(struct GstGameStr *s, struct GstActionStr *a){
  int i,d,i0,d0,i1,d1;

  // set search offset randomly
  i0=random()%8; 
  d0=random()%4;

  // search feasible action in ring order with offset
  for (i=0;i<8;i++){
    i1=(i+i0)%8; 
    a->ghost = &(s->ghost[GstPly0][i1]);
    for (d=0;d<4;d++){
      d1=(d+d0)%4;
      a->dir=d1;
      if (GstCheckAction(s,a)) return TRUE;
    }
  }

  a->ghost=NULL;
  a->dir=-1;
  return FALSE;
}

/*----------------------------------------------------------------*/
int main(int argc, char **argv) {
  char *hostaddr=NULL, *hostport=NULL;
  int waittime;
  struct GstGameStr gamestate, *s;
  struct GstActionStr action, *a;
  int r;

  // process arguments
  printf("usage: %s [hostaddr] [hostport] [waittime]\n", argv[0]);
  if (argc<3) {
    fprintf(stderr,"Please give host addr and player turn\n");
    exit(1);
  }
  hostaddr = argv[1];
  hostport = argv[2];
  printf("host = %s:%s\n",hostaddr,hostport);
  waittime=(argc>=3)?atoi(argv[3]):0;
  if (waittime<0) waittime=0;
  printf("wait = %dsec\n",waittime);

  // initialize the envirionments
  s=&gamestate;
  a=&action;
  {time_t t;time(&t);srandom(t);}//set ramdom seed with current time
  if (!GstLogOpen()) GstDie("unable to open log file");
  r=GstConnectHost(hostaddr,hostport);
  if (!r) GstDie("unable to connect host");

  // example for logging
  fprintf(gstlog,"Rits Geister\n");

  // starting the game
  GstInitGame(s);
  GstSetGhost(s);
  GstPrintGame(stdout,s);
  GstSendGhostInfo(s);

  sleep(waittime);
  s->turn=(strncmp(hostport,"10000",5)==0)?0:1;

  while(1){ // main loop
    r=GstRecvBoardState(s);
    GstPrintGame(stdout,s);

    if (r!=GstContinue) break;

    GstPlay(s,a);
    GstSendAction(a);
    GstExecAction(s,a);
    GstPrintGame(stdout,s);

    sleep(waittime);
  }

  if (r==GstWin) 
    printf("I won the game!\n");
  else if (r==GstLose)
    printf("I lost the game!\n");
  else 
    printf("What's happen?\n");

  GstCloseHost();
  GstLogClose();

  return 0;
}

/*----------------------------------------------------------------*/
void GstDie(char *msg){
  fprintf(stderr,"%s\n",msg);
  GstCloseHost();
  GstLogClose();
  exit(1);
}
