/*
  (C)2019 izumi@ieee.org, all rights reserved
*/

#define GEISTERLIB_C

#include "geisterlib.h"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define DEBUG 0

/*----------------------------------------------------------------*/
int GstLogOpen(void){
  time_t t;
  struct tm *d;
  char filename[256];

  time(&t);
  d=localtime(&t);
  sprintf(filename,"gstlog_%02d%02d%02d_%02d%02d%02d.txt",
	  d->tm_year-100,d->tm_mon+1,d->tm_mday,
	  d->tm_hour,d->tm_min,d->tm_sec);
  gstlog=fopen(filename,"w");
  if (gstlog==NULL) {
    fprintf(stderr,"unable to open log file (%s)\n",filename);
    return FALSE;
  }
  fprintf(gstlog,"geister log\n20%02d/%02d/%02d %02d:%02d:%02d\n",
	  d->tm_year-100,d->tm_mon+1,d->tm_mday,
	  d->tm_hour,d->tm_min,d->tm_sec);

  return TRUE;
}

int GstLogClose(void){
  fclose(gstlog);
  gstlog=NULL;
  return TRUE;
}

/*----------------------------------------------------------------*/
int GstInitGame(struct GstGameStr *s){
  int x,y,p,c,i;
  struct GstGhostStr *g;
  // initial positions of the ghosts
  const char ghost0x[2][8]={{1,2,3,4,1,2,3,4},{4,3,2,1,4,3,2,1}};
  const char ghost0y[2][8]={{4,4,4,4,5,5,5,5},{1,1,1,1,0,0,0,0}};

  if (DEBUG>0) fprintf(stderr,"GstInitGame()\n");

  // initialize the turn
  s->turn=0;

  // cleanup the board
  for (y=0;y<GstBdH;y++)
    for (x=0;x<GstBdW;x++)
      s->board[x][y]=NULL;

  // initialize the ghosts
  for (p=0;p<2;p++) { // 2players
    for (i=0;i<8;i++) { // 8ghosts
      g=&(s->ghost[p][i]);
      g->player=p;
      g->id=i;
      g->color=GstUnknown;
      g->state=GstOnboard;
      g->x=ghost0x[p][i];
      g->y=ghost0y[p][i];
      s->board[(int)g->x][(int)g->y]=g;
    }
  }

  // initialize the counter of captured ghosts
  for (p=0;p<2;p++) // 2players
    for (c=0;c<2;c++) // 2colors
      s->capturecount[p][c]=0;

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstCopyGame(struct GstGameStr *dst, struct GstGameStr *src, int reverse){
  int x,y,p,c,i;
  struct GstGhostStr *gdst,*gsrc;

  if (DEBUG>8) fprintf(stderr,"GstCopyGame()\n");
  if (DEBUG>0) GstCheckGame(src);

  dst->turn = src->turn;

  if (reverse==0) {
    for (p=0;p<2;p++)
      for (c=0;c<2;c++)
	dst->capturecount[p][c] = src->capturecount[p][c];
    for (p=0;p<2;p++)
      for (i=0;i<8;i++) {
	gdst = &(dst->ghost[p][i]);
	gsrc = &(src->ghost[p][i]);
	gdst->player = gsrc->player;
	gdst->id     = gsrc->id;
	gdst->color  = gsrc->color;
	gdst->state  = gsrc->state;
	gdst->x      = gsrc->x;
	gdst->y      = gsrc->y;
      }
  } else { // reverse the player and rotate the board
    for (p=0;p<2;p++)
      for (c=0;c<2;c++)
	dst->capturecount[p][c] = src->capturecount[GstOppositePlayer(p)][c];
    for (p=0;p<2;p++)
      for (i=0;i<8;i++) {
	gdst = &(dst->ghost[p][i]);
	gsrc = &(src->ghost[p][i]);
	gdst->player = GstOppositePlayer(gsrc->player);
	gdst->id     = gsrc->id;
	gdst->color  = gsrc->color;
	gdst->state  = gsrc->state;
	gdst->x      = 5 - gsrc->x;
	gdst->y      = 5 - gsrc->y;
      }
  }

  // cleanup the board
  for (y=0;y<GstBdH;y++)
    for (x=0;x<GstBdW;x++)
      dst->board[x][y]=NULL;

  // place the ghosts on the board
  for (p=0;p<2;p++)
    for (i=0;i<8;i++) {
      gdst = &(dst->ghost[p][i]);
      x=gdst->x;
      y=gdst->y;
      dst->board[x][y]=gdst;/* BUG! out of the board for captured BUG! */
    }

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstPrintGhostChar(FILE *fp, struct GstGhostStr *g){
  char c1,c2;

  if (g==NULL) {
    fprintf(fp,"  ");
    return TRUE;
  }
  
  c1='A'+g->id;
  c2=(g->color==GstRed )?'R'
    :(g->color==GstBlue)?'B'
    :'U';

  if (g->player==GstPly1) {
    c1=c1-'A'+'a'; // uppercase -> lowercase
    c2=c2-'A'+'a'; //  uppercase -> lowercase
  }

  switch (g->color) {
  case GstRed:  fprintf(fp,"\33[41m%c%c\33[0m",c1,c2);break;
  case GstBlue: fprintf(fp,"\33[44m%c%c\33[0m",c1,c2);break;
  default:      fprintf(fp,"\33[40m%c%c\33[0m",c1,c2);break;
  }

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstPrintBoard(FILE *fp, struct GstGameStr *s){
  int x,y;
  struct GstGhostStr *g;

  for (y=0;y<GstBdH;y++){
    fprintf(fp,"\t+--+--+--+--+--+--+\n\t");
    for (x=0;x<GstBdW;x++){
      g=s->board[x][y];
      fprintf(fp,"|");
      GstPrintGhostChar(fp,g);
    }
    fprintf(fp,"|\n");
  }
  fprintf(fp,"\t+--+--+--+--+--+--+\n");

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstPrintGame(FILE *fp, struct GstGameStr *s){
  int p,i;
  struct GstGhostStr *g;

  if (DEBUG>2) fprintf(stderr,"GstPrintState()\n");
  if (DEBUG>0) GstCheckGame(s);

  fprintf(fp,"----------------------------------------\n");
  fprintf(fp,"turn = %d\n",s->turn);
  fprintf(fp,"captured(0R,0B,1R,1B) = (%d,%d,%d,%d)\n",
	  s->capturecount[GstPly0][GstRed ],
	  s->capturecount[GstPly0][GstBlue],
	  s->capturecount[GstPly1][GstRed ],
	  s->capturecount[GstPly1][GstBlue]);

  for (p=0;p<2;p++)
    for (i=0;i<8;i++) {
      g=&(s->ghost[p][i]);
      if (g->state==GstEscaped) {
	GstPrintGhostChar(fp,g);
	fprintf(fp," escaped \n");
      }
    }

  fprintf(fp,"\t");
  p=GstPly0;
  for (i=0;i<8;i++) {
    g=&(s->ghost[p][i]);
    if (g->state==GstCaptured) {
      fprintf(fp," ");
      GstPrintGhostChar(fp,g);
    }
  }
  fprintf(fp,"\n");

  GstPrintBoard(fp,s);

  fprintf(fp,"\t");
  p=GstPly1;
  for (i=0;i<8;i++) {
    g=&(s->ghost[p][i]);
    if (g->state==GstCaptured) {
      fprintf(fp," ");
      GstPrintGhostChar(fp,g);
    }
  }
  fprintf(fp,"\n");

  fprintf(fp,"\n");

  return TRUE;
}

/*----------------------------------------------------------------*/
int GstApplyAction(struct GstGameStr *s, struct GstActionStr *a, int exec){
  int p,c,x,y,dir,x2,y2;
  struct GstGhostStr *g,*g2;

  if (DEBUG>32) fprintf(stderr,"GstApplyAction()\n");
  if (DEBUG>0) GstCheckGame(s);

  if (DEBUG>0) {
    if (s==NULL) GstDie("GstApplyAction: state is NULL\n");
    if (a==NULL) GstDie("GstApplyAction: action is NULL\n");
    if (a->ghost==NULL) GstDie("GstApplyAction: ghost is NULL\n");
  }

  dir=a->dir;
  g=a->ghost;
  p=g->player;
  c=g->color;
  x=g->x;
  y=g->y;

  if (dir<0 || dir>3) return FALSE;
  if (g->state!=GstOnboard) return FALSE;

  if ((c==GstBlue && p==GstPly0 && y==0 &&
       ((x==0 && (dir==GstWest || dir==GstNorth))||
	(x==5 && (dir==GstEast || dir==GstNorth)))) ||
      (c==GstBlue && p==GstPly1 && y==5 &&
       ((x==0 && (dir==GstWest || dir==GstSouth))||
	(x==5 && (dir==GstEast || dir==GstSouth))))){
    if (exec) {
      // ghost escapes
      s->board[x][y]=NULL;
      g->state=GstEscaped;
      g->x=-1;
      g->y=-1;
      s->turn+=1;
    }
    return TRUE;
  }

  x2=x+dir2dx[dir];
  y2=y+dir2dy[dir];

  // move out of the board
  if (x2<0 || x2>5 || y2<0 || y2>5) return FALSE;

  g2=s->board[x2][y2];
  if (g2==NULL) {
    if (exec) {
      // move to empty tile
      s->board[x][y]=NULL;
      s->board[x2][y2]=g;
      g->x=x2;
      g->y=y2;
      s->turn+=1;
    }
    return TRUE;
  } else if (g2->player==p) {
    // move on own ghost
    return FALSE;
  } else {
    if (exec) {
      // move and capture opponent ghost
      s->board[x][y]=NULL;
      s->board[x2][y2]=g;
      g->x=x2;
      g->y=y2;
      g2->state=GstCaptured;
      g2->x=-1;
      g2->y=-1;
      s->turn+=1;
    }
    return TRUE;
  }

  return FALSE;
}

/*----------------------------------------------------------------*/
int GstJudgeGame(struct GstGameStr *s){
  // judge the winning conditions
  int p,i;
  int r[2];

  if (DEBUG>16) fprintf(stderr,"GstJudgeGame()\n");
  if (DEBUG>0) GstCheckGame(s);

  for (p=0;p<2;p++)
    r[p]=0;

  for (p=0;p<2;p++)
    if (s->capturecount[p][GstRed ]==4)
      r[p]++;

  for (p=0;p<2;p++)
    if (s->capturecount[p][GstBlue]==4)
      r[GstOppositePlayer(p)]++;

  for (p=0;p<2;p++)
    for (i=0;i<8;i++)
      if (s->ghost[p][i].state==GstEscaped)
	r[p]++;

  if (r[GstPly0]==0 && r[GstPly1]==0) return GstContinue;
  if (r[GstPly0]==1 && r[GstPly1]==0) return GstWin;
  if (r[GstPly0]==0 && r[GstPly1]==1) return GstLose;

  GstDie("GstJudge: multiple winners\n");

  return GstContinue;
}

/*----------------------------------------------------------------*/
int GstCheckGame(struct GstGameStr *s){
  struct GstGhostStr *g;
  int x,y,p,c,i;
  int capture[2][2], color[2][2];

  if (DEBUG>32) fprintf(stderr,"GstCheckGame()\n");

  if (s==NULL) GstDie("game state is NULL");

  // check turn
  if (s->turn<0) GstDie("turn<0");

  // clear counts
  for (p=0;p<0;p++) {
    for (c=0;c<2;c++) {
      capture[p][c]=0;
      color[p][c]=0;
    }
  }

  // check ghosts
  for (p=0;p<0;p++) {
    for (i=0;i<8;i++) {
      g=&(s->ghost[p][i]);
      x=g->x;
      y=g->y;
      c=g->color;
      if (g->id!=i) GstDie("ghost id mismatch");
      switch (c) {
      case GstRed:
      case GstBlue:
	color[p][c]++;
      case GstUnknown:
	break;
      default:
	GstDie("ghost color error");
      }
      switch (g->state) {
      case GstOnboard:
	if (x<0 || x>5 || y<0 || y>5)
	  GstDie("onboard ghost is out of board");
	if (s->board[x][y]!=g)
	  GstDie("ghost position mismatch");
	break;
      case GstCaptured:
	capture[p][c]++;
	break;
      case GstEscaped:
	if (c==GstRed)
	  GstDie("red ghost escaped");
	break;
      default:
	GstDie("ghost state error");
      }
    }
  }

  // check counts
  for (p=0;p<0;p++) {
    for (c=0;c<2;c++) {
      if (capture[p][c]!=s->capturecount[p][c])
	GstDie("capture count mismatch");
      if (color[p][c]>4)
	GstDie("too many ghosts of a color");
    }
  }

  // check board
  for (y=0;y<GstBdH;y++)
    for (x=0;x<GstBdW;x++)
      if ((g=s->board[x][y])!=NULL)
	if (g->state!=GstOnboard || g->x!=x || g->y!=y)
	  GstDie("board mismatch");

  return TRUE;
}

/*----------------------------------------------------------------*/
int WeightedRandom(float w[], int n){
  int i;
  float sum,r;

  for (i=0,sum=0.0;i<n;i++)
    sum+=w[i];

  r=(random()%1024)/1024.0*sum;

  for (i=0,sum=0.0;i<n;i++) {
    sum+=w[i];
    if (r<sum) break;
  }

  return i;
}
